/* wydawca - automatic release submission daemon
   Copyright (C) 2007, 2008 Sergey Poznyakoff

   Wydawca is free software; you can redistribute it and/or modify it
   under the terms of the GNU General Public License as published by the
   Free Software Foundation; either version 3 of the License, or (at your
   option) any later version.

   Wydawca is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License along
   with wydawca. If not, see <http://www.gnu.org/licenses/>. */

#include "wydawca.h"
#include <mail.h>

int mailer_opened;
mu_mailer_t mailer;
mu_address_t admin_address;
mu_address_t from_address;
unsigned long mail_admin_mask;
struct message_template *admin_stat_message_template;
unsigned long owner_notification_flags;
char *user_message_template[MAX_EVENT];

void
mail_init ()
{
  if (!mailer)
    {
      int rc;
      if ((rc = mu_mailer_create (&mailer, NULL)))
	{
	  const char *url = NULL;
	  mu_mailer_get_url_default (&url);
	  logmsg (LOG_ERR, "cannot create default mailer `%s': %s",
		  url, mu_strerror (rc));
	}
    }
}

void
mail_send_message (mu_address_t rcpt, const char *text)
{
  int rc;
  mu_message_t msg;
  mu_stream_t stream = NULL;
  mu_header_t hdr;
  int mailer_flags = 0;
  static char *x_mailer = "wydawca (" PACKAGE_STRING ")";
  size_t size;
  char *buf;

  mu_message_create (&msg, NULL);

  mu_message_get_stream (msg, &stream);
  mu_stream_write (stream, text, strlen (text), 0, NULL);
  mu_message_get_header (msg, &hdr);
  mu_header_append (hdr, "X-Mailer", x_mailer);

  mu_address_to_string (rcpt, NULL, 0, &size);
  buf = xmalloc (size + 1);
  mu_address_to_string (rcpt, buf, size + 1, NULL);

  mu_header_set_value (hdr, "To", buf, 1);
  free (buf);
  
  if (debug_level > 1)
    {
      mu_debug_t debug;
      mu_mailer_get_debug (mailer, &debug);
      mu_debug_set_level (debug, MU_DEBUG_TRACE | MU_DEBUG_PROT);
      if (debug_level > 2)
	mailer_flags = MAILER_FLAG_DEBUG_DATA;
    }

  if (!mailer_opened)
    {
      if ((rc = mu_mailer_open (mailer, mailer_flags)))
	{
	  const char *url = NULL;
	  mu_mailer_get_url_default (&url);
	  logmsg (LOG_CRIT, "opening mailer `%s' failed: %s",
		  url, mu_strerror (rc));
	  return;
	}
      mailer_opened = 1;
    }
  
  rc = mu_mailer_send_message (mailer, msg, from_address, rcpt);
  if (rc)
    logmsg (LOG_CRIT, "cannot send message: %s", mu_strerror (rc));

  mu_message_destroy (&msg, mu_message_get_owner (msg));
}

void
mail_finish ()
{
  if (mailer_opened)
    mu_mailer_close (mailer);
}  


struct message_template
{
  struct message_template *next;
  char *name;
  char *text;
  /*  int mime; for future use */
};

static struct message_template *message_list; 

void
register_message_template (char *name, char *text)
{
  struct message_template *newp = xmalloc (sizeof *newp);
  newp->name = name;
  newp->text = text;
  newp->next = message_list;
  message_list = newp;
}

struct message_template *
find_message_template (char *name)
{
  struct message_template *p;

  for (p = message_list; p; p = p->next)
    if (strcmp (p->name, name) == 0)
      break;
  return p;
}

  
struct notification
{
  struct notification *next;
  enum notification_event ev;
  enum notification_target tgt;
  struct message_template *msg;
};

static struct notification *notification_list;

void
register_notification (enum notification_event ev,
		       enum notification_target tgt,
		       struct message_template *msg)
{
  struct notification *newp = xmalloc (sizeof *newp);
  newp->ev = ev;
  newp->tgt = tgt;
  newp->msg = msg;
  newp->next = notification_list;
  notification_list = newp;
}


void
mail_stats ()
{
  struct kw_expansion exp[MAX_STAT + 1];
  time_t t;
  char *text;
  
  if (!stat_mask_p (mail_admin_mask) || !mailer)
    return;

  if (debug_level)
    {
      size_t size;
      char *buf;
      mu_address_to_string (admin_address, NULL, 0, &size);
      buf = xmalloc (size + 1);
      mu_address_to_string (admin_address, buf, size + 1, NULL);
      logmsg (LOG_DEBUG, "Sending stats to %s", buf);
      free (buf);
    }
  
  if (dry_run_mode)
    return;

  time (&t);
  exp[0].kw = "date";
  exp[0].value = xstrdup (ctime (&t));
  exp[0].value [strlen (exp[0].value) - 1] = 0;
  exp[0].expand = NULL;
  
  make_stat_expansion (exp + 1);

  text = expand_param (admin_stat_message_template->text,
		       exp, sizeof (exp) / sizeof (exp[0]), NULL);

  mail_send_message (admin_address, text);

  free (text);
  free_kwexp (exp, sizeof (exp) / sizeof (exp[0]));
}

mu_address_t
get_recipient (struct access_method *method, struct file_triplet *trp,
	       char **errp)
{
  unsigned nrows, ncols, i;
  struct kw_expansion kwexp[4];
  mu_address_t rcpt = NULL;
  char *text;
  int rc;
  
  if (method->type == method_none)
    {
      *errp = "access method is not configured";
      return NULL;
    }

  if (method_init (method))
    {
      *errp = "failed to initialize access method";
      return NULL;
    }

  make_default_kwexp (kwexp, trp->user, trp->project);
  escape_kwexp (method, kwexp, NITEMS (kwexp));
  text = expand_param (method->param[1], kwexp, NITEMS (kwexp), NULL);

  rc = method_run (method, text);
  free (text);
  if (rc)
    {
      *errp = "cannot obtain recipient emails";
      return NULL;
    }
  
  nrows = method_num_rows (method);
  ncols = method_num_cols (method);

  if (nrows == 0)
    {
      *errp = "cannot obtain recipient emails";
      return NULL;
    }
  
  for (i = 0; i < nrows; i++)
    {
      mu_address_t addr;
      const char *str = method_result (method, i, 0);
      if (mu_address_create (&addr, str))
	continue;
      if (ncols > 0)
	{
	  str = method_result (method, i, 1);
	  if (str)
	    mu_address_set_personal (addr, 1, str);
	}
      mu_address_union (&rcpt, addr);
      mu_address_destroy (&addr);
    }

  return rcpt;
}

void
do_notify (struct file_triplet *trp, enum notification_event ev,
	   struct notification *ntf)
{
  mu_address_t rcpt = NULL;
  char *errp;
  switch (ntf->tgt)
    {
    case notify_admin:
      rcpt = admin_address;
      break;
      
    case notify_user:
      rcpt = get_recipient (trp->dpair->user_data_method, trp, &errp);
      break;
      
    case notify_owner:
      rcpt = get_recipient (trp->dpair->project_owner_method, trp, &errp);
    }

  if (!rcpt)
    {
      logmsg (LOG_ERR, "not notifying %s (project %s) about %s: %s",
	      notification_target_str (ntf->tgt),
	      trp->project,
	      notification_event_str (ev), errp);
      return;
    }

  if (debug_level)
    {
      size_t size;
      char *buf;
      mu_address_to_string (rcpt, NULL, 0, &size);
      buf = xmalloc (size + 1);
      mu_address_to_string (rcpt, buf, size + 1, NULL);
      logmsg (LOG_DEBUG, "notifying %s (project %s) about %s",
	      buf, trp->project, notification_event_str (ev));
      free (buf);
    }

  if (!dry_run_mode)
    {
      char *text = triplet_expand_param (ntf->msg->text, trp);
      mail_send_message (rcpt, text);
      free (text);
    }
  
  mu_address_destroy (&rcpt);
}

void
notify (struct file_triplet *trp, enum notification_event ev)
{
  struct notification *p;
  
  fill_project_name (trp);
  for (p = notification_list; p; p = p->next)
    if (p->ev == ev)
      do_notify (trp, ev, p);
  /* FIXME */
}
