/* wydawca - automatic release submission daemon
   Copyright (C) 2007 Sergey Poznyakoff

   Wydawca is free software; you can redistribute it and/or modify it
   under the terms of the GNU General Public License as published by the
   Free Software Foundation; either version 3 of the License, or (at your
   option) any later version.

   Wydawca is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License along
   with wydawca. If not, see <http://www.gnu.org/licenses/>. */

#include "wydawca.h"
#include "mail.h"

char *conffile = SYSCONFDIR "/wydawca.rc" ;
int debug_level;
int dry_run_mode;
int log_to_stderr = -1; /* -1 means autodetect */
int log_facility = LOG_LOCAL1;
char *syslog_tag = "wydawca";
int syslog_include_prio; /* syslog messages include priority */
unsigned long print_stats;         /* Print final statistics output */
time_t file_sweep_time = 0;
char *tar_command_name = "tar";
int enable_sendfile = 1;  /* Use sendfile by default */
			    
unsigned wydawca_stat[MAX_STAT];

void
usage ()
{
  printf ("usage: wydawca [OPTIONS]\n");
  printf ("wydawca synchronizes files from a set of upload directories with the\n"
	  "corresponding distribution sites.\n");
  printf ("\n");
  printf ("OPTIONS are:\n");
  printf ("  -c, --config-file=FILE         use FILE instead of the default configuration\n"
	  "                                 file\n");
  printf ("      --cron, --syslog           log to syslog\n");
  printf ("  -d, --debug                    increase debugging level by 1\n");
  printf ("  -e, --stderr                   log to stderr\n");
  printf ("  -I, --include-directory[=DIR]  set default include directory\n");
  printf ("  -n, --dry-run                  do nothing, print almost everything;\n"
	  "                                 implies `--debug --stderr'; use additional\n"
	  "                                 `--debug' options to get even more info\n");
  printf ("  -t, --lint                     parse configuration file and exit\n");
  printf ("  -h, --help                     print this help list\n");
  printf ("  -v, --version                  print program version and exit\n");

  printf ("\nThe defaults are:\n");
  printf ("Configuration file: %s\n", conffile);
  printf ("Syslog facility: %s\n", gsc_syslog_facility_to_str (log_facility));
  printf ("Syslog tag: %s\n", syslog_tag);
  printf ("\n");
  printf ("Report bugs to <%s>\n", PACKAGE_BUGREPORT);
}


/* Logging */
void
syslog_printer (int prio, const char *fmt, va_list ap)
{
  if (syslog_include_prio)
    {
      static char *fmtbuf;
      static size_t fmtsize;
      const char *p = gsc_syslog_priority_to_str (prio);
      size_t size = strlen (p) + 3 + strlen (fmt) + 1;

      if (size > fmtsize)
	{
	  fmtsize = size;
	  fmtbuf = x2realloc (fmtbuf, &fmtsize);
	}
      sprintf (fmtbuf, "[%s] %s", p, fmt);
      fmt = fmtbuf;
    }
  
#if HAVE_VSYSLOG
  vsyslog (prio, fmt, ap);
#else
  char buf[128];
  vsnprintf (buf, sizeof buf, fmt, ap);
  syslog (prio, "%s", buf);
#endif
}

void
stderr_printer (int prio, const char *fmt, va_list ap)
{
  const char *p = gsc_syslog_priority_to_str (prio);
  fprintf (stderr, "%s: ", program_name);

  if (p)
    fprintf (stderr, "[%s] ", p);

  vfprintf (stderr, fmt, ap);
  fputc ('\n', stderr);
}


static void (*log_printer) (int prio, const char *fmt, va_list ap) =
         stderr_printer;

void
logmsg (int prio, char *fmt, ...)
{
  va_list ap;

  switch (prio)
    {
    case LOG_EMERG:
    case LOG_ALERT:
    case LOG_CRIT:
    case LOG_ERR:
      UPDATE_STATS (STAT_ERRORS);
      break;
      
    case LOG_WARNING:
      UPDATE_STATS (STAT_WARNINGS);
    }
      
  va_start (ap, fmt);
  log_printer (prio, fmt, ap);
  va_end (ap);
}



enum {
  SYSLOG_OPTION = 256
};

struct option options[] = {
  { "config-file", required_argument, NULL, 'c' },
  { "debug", no_argument, NULL, 'd' },
  { "stderr", no_argument, NULL, 'e' },
  { "syslog", no_argument, NULL, SYSLOG_OPTION },
  { "cron", no_argument, NULL, SYSLOG_OPTION },
  { "help", no_argument, NULL, 'h' },
  { "version", no_argument, NULL, 'v' },
  { "dry-run", no_argument, NULL, 'n' },
  { "lint", no_argument, NULL, 't' },
  { "include-directory", optional_argument, NULL, 'I' },
  { NULL }
};


static char *stat_name[MAX_STAT] = {
  "errors",
  "warnings",
  "bad signatures",
  "access violation attempts",
  "complete triplets",
  "incomplete triplets",
  "bad triplets",
  "expired triplets",
  "triplet successes",
  "files uploaded",
  "files archived",
  "symlinks created",
  "symlinks removed",
};

static char *stat_kwname[MAX_STAT] = {
  "stat:errors",               
  "stat:warnings",             
  "stat:bad_signatures",       
  "stat:access_violations",    
  "stat:complete_triplets",    
  "stat:incomplete_triplets",  
  "stat:bad_triplets",         
  "stat:expired_triplets",     
  "stat:triplet_success",      
  "stat:uploads",              
  "stat:archives",             
  "stat:symlinks",             
  "stat:rmsymlinks",      
};

int
stat_mask_p (unsigned long mask)
{
  int i;

  for (i = 0; i < MAX_STAT; i++)
    if (wydawca_stat[i] != 0 && (mask && STAT_MASK (i)))
      return 1;
  return 0;
}

static char *
stat_expand (struct kw_expansion *exp, void *data)
{
  char sbuf[INT_BUFSIZE_BOUND (wydawca_stat[0])];
  exp->value = xstrdup (uinttostr (wydawca_stat[(int) exp->data], sbuf));
  return exp->value;
}

void
make_stat_expansion (struct kw_expansion *exp)
{
  int i;
  for (i = 0; i < MAX_STAT; i++)
    {
      exp[i].kw = stat_kwname[i];
      exp[i].value = NULL;
      exp[i].static_p = 0;
      exp[i].expand = stat_expand;
      exp[i].data = (void*) i;
    }
}

void
logstats ()
{
  int i;

  if (stat_mask_p (print_stats))
    {
      for (i = 0; i < MAX_STAT; i++)
	if (print_stats & STAT_MASK (i))
	  logmsg (LOG_INFO, "%s: %u", stat_name[i], wydawca_stat[i]);
    }
  
  mail_stats ();
}



int
main (int argc, char **argv)
{
  int c;
  int lint_mode = 0;
  
  program_name = argv[0];
  gsc_config_include_dir = SYSCONFDIR;
  mu_register_all_mailer_formats ();
  while ((c = getopt_long (argc, argv, "c:dehI::ntv", options, NULL)) != EOF)
    {
      switch (c)
	{
	case 'c':
	  conffile = optarg;
	  break;

	case 'd':
	  debug_level++;
	  break;
	  
	case 'e':
	  log_to_stderr = 1;
	  break;
	  
	case SYSLOG_OPTION:
	  log_to_stderr = 0;
	  break;
	  
	case 'h':
	  usage ();
	  exit (0);

	case 'I':
	  gsc_config_include_dir = optarg;
	  break;
	  
	case 'n':
	  log_to_stderr = 1;
	  debug_level = 1;
	  dry_run_mode = 1;
	  break;

	case 't':
	  lint_mode = 1;
	  log_to_stderr = 1;
	  break;
	  
	case 'v':
	  gsc_version ("wydawca");
	  exit (0);

	default:
	  error (1, 0, "unknown option: %c", c);
	}
    }

  if (argc != optind)
    error (1, 0, "extra command line arguments");

  parse_config ();
  if (lint_mode)
    exit (0);

  if (log_to_stderr == -1)
    log_to_stderr = isatty (0);
  
  if (!log_to_stderr)
    {
      openlog (syslog_tag, LOG_PID, log_facility);
      log_printer = syslog_printer;
    }

  mail_init ();
  
  logmsg (LOG_NOTICE, "wydawca (%s) started", PACKAGE_STRING);
  scan_directories ();

  logmsg (LOG_NOTICE, "wydawca (%s) finished", PACKAGE_STRING);
  logstats ();

  mail_finish ();

  exit (0);
}
